------------------------------------------------------------------------------
--				Highlight code for plenty of conditions						--
--	supports Buffs, Debuffs, Equiped items, out of mana, no range,			--
--	not usable, currently casting, currently channeling and autoattacking	--
------------------------------------------------------------------------------
local Highlighting =  FlexBar2:NewModule("Highlighting");
-- If this var is incremented, SetDefaults() will be called for this module thus resetting all its settings. Though there are no settings for this module, it has to be declared
Highlighting.Version = 1;
-- Get a local version of SpecialEvents-Aura and PaintChips-2.0
local Aura = AceLibrary("SpecialEvents-Aura-2.0");
local PaintChips = AceLibrary("PaintChips-2.0");

-- Note: Both Inner & Outer Priority will be configurable in the future
-- Define priority for inner checks (mana, range & usable)
local InnerPriority = {"NoRange", "NoMana", "Unusable"};
-- Define priority for outer checks (channeling, casting, autoattacking, buffed, debuffed, equipped)
local OuterPriority = {"Channeling", "Casting","AutoCasting", "Buff", "Debuff", "Equipped"};
local RangeGuess;

-- Register all colors with paintchips, scheduled to be configurable in the future
function Highlighting:OnInitialize()
    PaintChips:RegisterColor("Buff", 0 , 255, 0);
    PaintChips:RegisterColor("Debuff", 255 , 0, 0);
    PaintChips:RegisterColor("Equipped", 255 , 255, 0);
    PaintChips:RegisterColor("NoMana", 0 , 0, 150);
    PaintChips:RegisterColor("NoRange", 150 , 0, 0);
    PaintChips:RegisterColor("Unusable", 100 , 100, 100);
    PaintChips:RegisterColor("Casting", 255, 255, 255);
    PaintChips:RegisterColor("Channeling", 148, 0, 211);
    PaintChips:RegisterColor("AutoCasting", 238, 118, 0);
end

-- Declare the button mixin
Highlighting.ButtonMixin = FlexBar2.OO.Mixin
{
    "SetOuterHighlight",
    "SetInnerHighlight",
    "UpdateOuterHighlight",
    "CheckCasting",
    "CheckChanneling",
    "CheckStartAutoCasting",
    "CheckStopAutoCasting",
    "CheckBuff",
    "CheckDebuff",
    "CheckUsable",
    "CheckEquipped",
    "CheckRange",
};

function Highlighting.ButtonMixin:Activate()
	-- Variables to store what checks are currently active
    self.InnerCheck = {};
    self.OuterCheck = {};
	-- Update the outer highlight when clicked, as clicking might check/uncheck it
    self:RegisterScript("PostClick", "UpdateOuterHighlight");
	-- Update range highlight every 0.3 sec with FB2_UPDATE
    self:RegisterEvent("FB2_UPDATE", "CheckRange");
	-- Update usability on mana, rage & energy updates (this includes mana)
    self:RegisterEvent("UNIT_MANA", function(Button, Event, Source, UnitId) if(UnitId == "player") then self:CheckUsable(); end end);
    self:RegisterEvent("UNIT_RAGE", function(Button, Event, Source, UnitId) if(UnitId == "player") then self:CheckUsable(); end end);
    self:RegisterEvent("UNIT_ENERGY", function(Button, Event, Source, UnitId) if(UnitId == "player") then self:CheckUsable(); end end);
	-- Update equiped items when inventory changes
    self:RegisterEvent("UNIT_INVENTORY_CHANGED", function(Button, Event, Source, UnitId) if(UnitId == "player") then self:CheckEquipped(); end end);
	-- Update casting on spellcasting start & stop
	-- Blizzard API is currently, buggy, using FB2_UPDATE till 2.3
	--[[
    self:RegisterEvent("UNIT_SPELLCAST_START", function(Button, Event, Source, UnitId) if(UnitId == "player") then self:CheckCasting(); end end);
    self:RegisterEvent("UNIT_SPELLCAST_STOP", function(Button, Event, Source, UnitId) if(UnitId == "player") then self:CheckCasting(); end end);
	self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED", function(Button, Event, Source, UnitId) if(UnitId == "player") then self:CheckCasting(); end end);
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", function(Button, Event, Source, UnitId) if(UnitId == "player") then self:CheckCasting(); end end);
	--]]
	self:RegisterEvent("FB2_UPDATE", "CheckCasting");
	-- Update channeling on channeling start & stop
    self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START", function(Button, Event, Source, UnitId) if(UnitId == "player") then self:CheckChanneling(); end end);
    self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP", function(Button, Event, Source, UnitId) if(UnitId == "player") then self:CheckChanneling(); end end);
	-- Update autoattacking (or wanding, shooting, whatever) with StartAutoRepeat and EnterCombat events
    self:RegisterEvent("START_AUTOREPEAT_SPELL", "CheckStartAutoCasting");
    self:RegisterEvent("STOP_AUTOREPEAT_SPELL", "CheckStopAutoCasting");
    self:RegisterEvent("PLAYER_ENTER_COMBAT", "CheckStartAutoCasting");
    self:RegisterEvent("PLAYER_LEAVE_COMBAT", "CheckStopAutoCasting");
	-- Update buffs when SpecialEvent triggers on player buffs
    self:RegisterEvent("SpecialEvents_PlayerBuffGained", "CheckBuff");
    self:RegisterEvent("SpecialEvents_PlayerBuffLost", "CheckBuff");
	-- Update usability when the player gains debuffs (hint: forbearance)
	self:RegisterEvent("SpecialEvents_PlayerDebuffGained", "CheckUsable");
	self:RegisterEvent("SpecialEvents_PlayerDebuffLost", "CheckUsable");
	-- Update debuffs when the target gains a debuff
    self:RegisterEvent("SpecialEvents_UnitDebuffGained", "CheckDebuff");
    self:RegisterEvent("SpecialEvents_UnitDebuffLost", "CheckDebuff");
    self:RegisterEvent("SpecialEvents_AuraTargetChanged", "CheckDebuff");
	-- Update everything if the action changed for this button
    self:RegisterEvent("FB2_ActionChanged", function(self, Event, Dispatcher, Button) if(Button == self) then self:CheckBuff(); end end);
    self:RegisterEvent("FB2_ActionChanged", function(self, Event, Dispatcher, Button) if(Button == self) then self:CheckDebuff(); end end);
    self:RegisterEvent("FB2_ActionChanged", function(self, Event, Dispatcher, Button) if(Button == self) then self:CheckChanneling(); end end);
    self:RegisterEvent("FB2_ActionChanged", function(self, Event, Dispatcher, Button) if(Button == self) then self:CheckCasting(); end end);
    self:RegisterEvent("FB2_ActionChanged", function(self, Event, Dispatcher, Button) if(Button == self) then self:CheckEquipped(); end end);
    self:RegisterEvent("FB2_ActionChanged", function(self, Event, Dispatcher, Button) if(Button == self) then self:CheckUsable(); end end);
	-- Highlight upon loading too
    self:CheckEquipped();
    self:CheckBuff();
    self:CheckDebuff();
    self:CheckUsable();
end

-- Make sure the outer highlight is set correctly, called on postclick and when a highlight changes
function Highlighting.ButtonMixin:UpdateOuterHighlight()
    if(type(self.OuterCheck[1]) == "string") then
		self.Frame:SetChecked(true);
		local _, r, g, b = PaintChips:GetRGBPercent(self.OuterCheck[1])
		self.Frame:GetCheckedTexture():SetVertexColor(r, g, b);
	else
		self.Frame:SetChecked(false);
	end
end

function Highlighting.ButtonMixin:SetInnerHighlight(CheckType, State)
	-- State is true, or an equivalent of true
	if(State) then
		-- Make sure this highlight is not already active
		if(not FlexBar2:ValueIsInTable(self.InnerCheck, CheckType)) then
			-- insert the highlight into the table
			table.insert(self.InnerCheck, CheckType);
			-- sort the table with the priorities in our priority tables declared at the top
			table.sort(self.InnerCheck, function(a, b)
				return FlexBar2:GetKeyByValue(InnerPriority, a) < FlexBar2:GetKeyByValue(InnerPriority, b);
			end);
		end
	-- State is false or an equivalent of false
	else
		-- Remove the highlight
		FlexBar2:RemoveByValue(self.InnerCheck, CheckType);
	end
	-- find out if there's still a highlight active after doing our stuff
	if(type(self.InnerCheck[1]) == "string") then
		-- If there's a highlight active, set it
		local _, r, g, b = PaintChips:GetRGBPercent(self.InnerCheck[1])
		self.Frame.Icon:SetVertexColor(r, g, b);
    else
		-- If there's no highlight active, back to plain boring white
        self.Frame.Icon:SetVertexColor(1, 1, 1);
	end
end

function Highlighting.ButtonMixin:SetOuterHighlight(CheckType, State)
	-- State is true, or an equivalent of true
	if(State) then
		-- Make sure this highlight is not already active
		if(not FlexBar2:ValueIsInTable(self.OuterCheck, CheckType)) then
			-- insert the highlight into the table
			table.insert(self.OuterCheck, CheckType);
			-- sort the table with the priorities in our priority tables declared at the top
			table.sort(self.OuterCheck, function(a, b)
				return FlexBar2:GetKeyByValue(OuterPriority, a) < FlexBar2:GetKeyByValue(OuterPriority, b);
			end);
		end
	else
		FlexBar2:RemoveByValue(self.OuterCheck, CheckType)
	end
	-- Update the outer highlight
	self:UpdateOuterHighlight();
end

-- Check if the user is currently auto casting something, be it autoattack, autowand or autoshot
function Highlighting.ButtonMixin:CheckStartAutoCasting()
    local Spell = self:GetModifiedAttribute("spell"); 
    if(Spell) then
        self:SetOuterHighlight("AutoCasting", IsAutoRepeatSpell(Spell) or IsAttackSpell(Spell));
    else
        self:SetOuterHighlight("AutoCasting", false);
    end
end

-- Triggered when autocasting stops
function Highlighting.ButtonMixin:CheckStopAutoCasting()
    self:SetOuterHighlight("AutoCasting", false);
end

-- Checks if the user is currently casting something
function Highlighting.ButtonMixin:CheckCasting()
    local SpellName, SpellRank = UnitCastingInfo("player");
    if(SpellName) then
        if(SpellRank) then
            SpellName = SpellName .. "(" .. SpellRank .. ")";
        end
        self:SetOuterHighlight("Casting", self:GetModifiedAttribute("spell") == SpellName);
    else
        self:SetOuterHighlight("Casting", false);
    end
	-- the IsCurrentSpell function used here comes out in 2.3
	--[[
	local ActionType = self:GetModifiedAttribute("type");
	if(ActionType == "spell") then
		self:SetOuterHighlight("Casting", IsCurrentSpell(self:GetModifiedAttribute("spell")));
	elseif(ActionType == "item") then
		self:SetOuterHighlight("Casting", IsCurrentItem(self:GetModifiedAttribute("itemlink")));
	elseif(ActionType == "macro") then
		-- Not supported yet
		self:SetOuterHighlight("Casting", false);
	else
		self:SetOuterHighlight("Casting", false);
	end-]]
end

-- Checks for channeled spells
function Highlighting.ButtonMixin:CheckChanneling()
    local SpellName, SpellRank = UnitCastingInfo("player");
    if(SpellName) then
        if(SpellRank) then
            SpellName = SpellName .. "(" .. SpellRank .. ")";
        end
        self:SetOuterHighlight("Channeling", self:GetModifiedAttribute("spell") == SpellName);
    else
        self:SetOuterHighlight("Channeling", false);
    end
end

-- Checks for buffs on the player
function Highlighting.ButtonMixin:CheckBuff()
    local Icon = self:GetIcon();
    if(Icon ~= nil and Aura:UnitHasBuff("player", nil, nil, Icon)) then
        self:SetOuterHighlight("Buff", true);
    else
        self:SetOuterHighlight("Buff", false);
    end
end

-- Checks for debuffs on the target
function Highlighting.ButtonMixin:CheckDebuff()
    if(self:GetIcon() ~= nil and Aura:UnitHasDebuff("target", nil, nil, self:GetIcon())) then
        self:SetOuterHighlight("Debuff", true);
    else
        self:SetOuterHighlight("Debuff", false);
    end
end

-- Check for equiped items
function Highlighting.ButtonMixin:CheckEquipped()
    local Item = self:GetModifiedAttribute("itemlink");
    if(Item and IsEquippedItem(Item)) then
        self:SetOuterHighlight("Equipped", true);
    else
        self:SetOuterHighlight("Equipped", false);
    end
end

-- Check if the action is in range
function Highlighting.ButtonMixin:CheckRange()
    local ButtonType = self:GetModifiedAttribute("type");
    local IsInRange;
    if(ButtonType == "spell") then
        local Spell = self:GetModifiedAttribute("spell");
        IsInRange = IsSpellInRange(Spell, "target");
    elseif(ButtonType == "item") then
        local Item = self:GetModifiedAttribute("itemlink");
        local IsInRange = IsItemInRange(Item, "target"); 
    else
        self:SetInnerHighlight("NoRange", false);
        return;
    end
    self:SetInnerHighlight("NoRange", IsInRange == 0);
end

-- CHeck if the action is usable 
function Highlighting.ButtonMixin:CheckUsable()
    local ButtonType = self:GetModifiedAttribute("type");
    local Usable, NoMana;
    if(ButtonType == "spell") then
        local Spell = self:GetModifiedAttribute("spell");
        Usable, NoMana = IsUsableSpell(Spell); 
    elseif(ButtonType == "item") then
        local Item = self:GetModifiedAttribute("itemlink");
        Usable, NoMana = IsUsableItem(Item);
    else
        self:SetInnerHighlight("NoMana", false);
        self:SetInnerHighlight("Unusable", false);
        return;
    end
    if(Usable) then
            self:SetInnerHighlight("Unusable", false);
            self:SetInnerHighlight("NoMana", false);
    else
        self:SetInnerHighlight("Unusable", true);
        if(NoMana) then
            self:SetInnerHighlight("NoMana", true);
        else
            self:SetInnerHighlight("NoMana", false);
        end
    end
end
